
function auto_align_images_func,pq

   common auto_align_images_private3,rimage,timage, $
                                    npq,rnx,rny,nxt,nyt, $
                                    restricted,quiet

   if restricted then begin
      tr = rss2pq(nxt,nyt,xshift=pq[3],yshift=pq[4],rot12=pq[0], $
                  xscale=pq[1],yscale=pq[2],/center,p=p,q=q)
   endif else begin
      p = pq[0L:npq-1L]
      q = pq[npq:*]
   endelse
   missing = long(min(timage)-1000.)
   itest = poly_2d(timage,p,q,1,rnx,rny,missing=missing)

   good = where(itest GT missing,ngood)
   if ngood GT 0 then begin
      x = good MOD rnx
      y = good / rnx
      bad = where(itest LE missing,nbad)
      if nbad GT 0 then begin  ; Zero out the missing points
         itest[bad] = 0.0
      endif
      minx = min(x) & maxx = max(x)
      miny = min(y) & maxy = max(y)
      snx = maxx-minx+1L
      sny = maxy-miny+1L
      ; Cut off zero areas to save computation time
      srimage = rimage[minx:maxx,miny:maxy]
      sitest = itest[minx:maxx,miny:maxy]
      nxy = n_elements(srimage)
      appodize = hanning(snx,sny)
      ; full correlation for testing
      ;;crosscorr = $
      ;;abs(fft(conj(fft((srimage-total(srimage)/nxy)*appodize,-1))*fft
      ;;((sitest-total(sitest)/nxy)*appodize,-1),1)) 
      ; The cross correlation at zero lag:
      crosscorr = total(((srimage-total(srimage)/nxy)*appodize)*  $
                        ((sitest -total(sitest) /nxy)*appodize))/nxy
      if NOT keyword_set(quiet) then begin
         if !d.x_size LT snx OR !d.y_size LT sny then begin
            if !d.window GE 0 then wdelete,!d.window
            wdef,ws,snx,sny
         endif
         erase
         tvscl,srimage
         ocontour2,sitest,level=max(sitest)*[0.1,0.25,0.50,0.75,0.90]
         xyouts,/normal,0.0,0.0,string(crosscorr[0,0]),align=0.0
      endif
      return,-crosscorr[0,0]  ; negative since powel minimizes, 
                              ; but we want to maximize the cross corr
   endif else return,0.0  ; No good points so must be WAY off.

end

;+

function auto_align_images,timagein,rimagein,pin,qin,pout,qout, $
                           unrestricted=unrestrictedin,quiet=quietin, $
                           missing=missval,itmax=itmaxin,double=double

;NAME:
;     AUTO_ALIGN_IMAGES
;PURPOSE:
;     Align arbitrary images using a cross correlation.  This routine is
;     particularly useful for images from different instruments in which
;     a very general transformation between the two images is required.  
;     Don't use this routine if you only need to shift the two images on
;     top of one another.  It'll work, but it will be much slower than
;     you need.
;CATEGORY:
;CALLING SEQUENCE:
;     inew=auto_align_images(transformed_image,reference_image, $
;                            pin,qin,pout,qout)
;INPUTS:
;     transformed_image = image to be transformed onto the reference image.
;     refererence_image = reference image
;     pin,qin = Initial guess for the p and q passed to poly_2d.  You must get
;               this pretty close!! e.g. you can use the mouse to click on 
;               corresponding points like this:
;           pp = setpts(transformed_image,reference_image)
;           tt = caltrans(pp)
;           pin = tt[*,0]
;           qin = tt[*,1]
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     missing = Value to use where the transformed image is out of the 
;               field of view of the initial image to be transformed.
;     /unrestricted = Do not restrict the transformation to rotation, shift, 
;                     and scale, i.e. allow a general very general warping.
;                     Chances are you do not want to set this keyword in most 
;                     circumstances.  You might use it if the images have
;                     atmospheric blurring or something like that.
;     itmax = maximum number of iterations (default = 200).
;     /double = use double precision.
;     /quiet = work quietly
;OUTPUTS:
;     inew = new transformed image overlaid in reference image.
;     pout,qout = Output transformation.  
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The initial guess (pin,qin) must be close.  This routine only refines
;     that guess.
;
;     The features to be matched should be generally towards the center of
;     the images and away from the edges, since the cross correlation uses a 
;     Hanning window.
;
;     *Very slow*
;PROCEDURE:
;
;     The initial warping is tweaked until a maximum in the cross correlation
;     is found using the powell minimization procedure.
;
;     The transformed image overlaid on the reference image is:
;        (where rnx,rny is the size of the reference image)
;
;     new_image = poly_2d(transformed_image,pout,qout,1,rnx,rny,missing=0.0)
;     tvscl,reference_image
;     ocontour2,new_image
;
;MODIFICATION HISTORY:
;     T. Metcalf 2001-09-25
;     T. Metcalf 2001-09-26 Improved the way the display windows work.  If the
;                           window is too small, it will appear the that 
;                           alignment is wrong even though it is right.  So,
;                           make sure that the windows are always large
;                           enough.
;-

common auto_align_images_private3,rimage,timage, $
                                 npq,rnx,rny,nxt,nyt, $
                                 restricted,quiet

if keyword_set(itmaxin) then itmax=itmaxin else itmax=200
if keyword_set(unrestrictedin) then restricted=0 else restricted=1
if keyword_set(quietin) then quiet=1 else quiet=0

rnx = n_elements(rimagein[*,0])
rny = n_elements(rimagein[0,*])
nxt = n_elements(timagein[*,0])
nyt = n_elements(timagein[0,*])

rimage = rimagein
timage = timagein

if restricted then begin
   if NOT quiet then begin
      print
      print,'Initial guess:'
   endif
   pq2rss,pin,qin,erot,exscl,eyscl,exshft,eyshft,enrss,nxt,nyt, $
          /center,quiet=quiet
   pq = [erot,exscl,eyscl,exshft,eyshft]
   npq = -1
endif else begin
   npq = n_elements(pin)
   if n_elements(qin) NE npq then $
      message,'p and q must have the same dimension'
   pq = [pin,qin]
endelse
npq2 = n_elements(pq)

xi = fltarr(npq2,npq2)
for i=0L,npq2-1L do xi[i,i]=1.0   ; unit vectors

; Now do the minimization to find the best alignment

powell,pq,xi,1.e-4,mincorr,'auto_align_images_func', $
       iter=niter,itmax=200,double=double

if niter GE itmax then message,/info,'WARNING: Failed to fully converge'

if restricted then begin
   tr = rss2pq(nxt,nyt,xshift=pq[3],yshift=pq[4],rot12=pq[0], $
               xscale=pq[1],yscale=pq[2],/center,p=p,q=q)
endif else begin
   p = pq[0L:npq-1L]
   q = pq[npq:*]
endelse

itest = poly_2d(timage,p,q,1,rnx,rny,missing=missval)
if NOT quiet then begin
   if !d.x_size LT rnx OR !d.y_size LT rny then wdef,ws,rnx,rny
   tvscl,rimage
   ocontour2,itest,level=max(itest)*[0.1,0.25,0.50,0.75,0.90]
endif

if restricted and NOT quiet then begin
   print
   print,'Final parameters:'
   pq2rss,p,q,erot,exscl,eyscl,exshft,eyshft,enrss,nxt,nyt,/center
end

pout = p
qout = q
return,itest

end
